<?php

declare(strict_types=1);

namespace Drupal\entity_sync_odata\EntitySync\Api;

// Drupal modules.
use Drupal\entity_sync\Client\ClientInterface;
use Drupal\entity_sync\Client\SupportsDeleteClientInterface;
// Third-party libraries.
use GuzzleHttp\HandlerStack;
use GuzzleHttp\Middleware;
use KrystalCode\Api\Session\SessionManagerInterface;
use KrystalCode\ApiIterator\ClientInterface as IteratorClientInterface;
use Psr\Http\Message\RequestInterface;

/**
 * The API client Entity Sync adapter for OData v4.0 APIs.
 */
class EntityClient implements ClientInterface, SupportsDeleteClientInterface {

  use ODataClientBuilderTrait;

  /**
   * Constructs a new EntityClient object.
   *
   * @param \KrystalCode\Api\Session\SessionManagerInterface $sessionManager
   *   The session manager.
   * @param \KrystalCode\ApiIterator\ClientInterface $iteratorClient
   *   The API iterator client.
   * @param string $entityType
   *   The entity type (resource) of the API client that will be used
   *   e.g. `People`.
   * @param array $oData
   *   An array containing `select` and `expand` items to add to the
   *   corresponding OData clauses. Currenlty only supported by the
   *   `importList` client method. OData filters are not supported yet.
   */
  public function __construct(
    protected SessionManagerInterface $sessionManager,
    protected IteratorClientInterface $iteratorClient,
    protected string $entityType,
    protected array $oData = []
  ) {
  }

  /**
   * {@inheritdoc}
   *
   * phpcs:disable
   * @I Document all supported client options
   *    type     : task
   *    priority : normal
   *    labels   : documentation
   * phpcs:enable
   */
  public function importList(array $filters = [], array $options = []) {
    if (!isset($options['odata'])) {
      $options['odata'] = $this->oData;
    }

    return $this->iteratorClient->list(
      $this->buildListRequestOptions($filters, $options),
      [
        'select' => $options['odata']['select'] ?? NULL,
        'filter' => [$filters, $options],
        'expand' => $options['odata']['expand'] ?? NULL,
      ]
    );
  }

  /**
   * {@inheritdoc}
   */
  public function importEntity($id) {
    return (object) $this
      ->getOdataClient()
      ->from($this->entityType)
      ->find($id)
      ->getProperties();
  }

  /**
   * {@inheritdoc}
   */
  public function create(array $fields, array $options = []) {
    // Add header to get the created entity in the response. It is needed by
    // managed synchronizations so that we update in Drupal the `changed` time
    // of the remote entity. The fields to be included in the response can be
    // limited by using the `$select` query parameter in the URL - not supported
    // yet to be provided via `$options`.
    // @I Support passing OData query parameters to `create` via options
    //    type     : improvement
    //    priority : normal
    //    labels   : export
    $stack = HandlerStack::create();
    $stack->push(Middleware::mapRequest(
      function (RequestInterface $request) {
        return $request->withHeader('Prefer', 'return=representation');
      })
    );

    // @I Review response to `POST` requests according to OData v4
    //    type     : bug
    //    priority : normal
    //    labels   : export
    //    notes    : In Microsoft Dataverse response the object is contained two
    //               levels in in the response. It is not clear though whether
    //               that is determined by OData or by the OData client, or
    //               specific to Microsoft Dataverse.
    return (object) $this
      ->getOdataClient(['handler' => $stack])
      ->post("{$this->entityType}", $fields)[0][0]
      ->getProperties();
  }

  /**
   * {@inheritdoc}
   */
  public function update($id, array $fields, array $options = []) {
    // Add header to get the updated entity in the response. It is needed by
    // managed synchronizations so that we update in Drupal the `changed` time
    // of the remote entity. The fields to be included in the response can be
    // limited by using the `$select` query parameter in the URL - not supported
    // yet to be provided via `$options`.
    // @I Support passing OData query parameters to `update` via options
    //    type     : improvement
    //    priority : normal
    //    labels   : export
    $stack = HandlerStack::create();
    $stack->push(Middleware::mapRequest(
      function (RequestInterface $request) {
        return $request->withHeader('Prefer', 'return=representation');
      })
    );

    // @I Review response to `PATCH` requests according to OData v4
    //    type     : bug
    //    priority : normal
    //    labels   : export
    //    notes    : In Microsoft Dataverse response the object is contained two
    //               levels in in the response. It is not clear though whether
    //               that is determined by OData or by the OData client, or
    //               specific to Microsoft Dataverse.
    return (object) $this
      ->getOdataClient(['handler' => $stack])
      ->patch("{$this->entityType}({$id})", $fields)[0][0]
      ->getProperties();
  }

  /**
   * {@inheritdoc}
   */
  public function delete($id, array $options = []) {
    // @I Review response to `DELETE` requests according to OData v4
    //    type     : bug
    //    priority : normal
    //    labels   : export
    //    notes    : We have not been able to fully test this but it seems that
    //               the response to deletions is 204 with empty response body.
    //               We should validate that this is what OData v4 prescribes.
    $this
      ->getOdataClient()
      ->delete("{$this->entityType}({$id})");

    // Nothing is contained in the response body, so no need to parse it. If the
    // resource was not found, exception would have been raised.
    return new \stdClass();
  }

  /**
   * Builds and returns the options for the API iterator client.
   *
   * @param array $filters
   *   The filters passed to the client.
   * @param array $options
   *   The options passed to the client.
   *
   * @return array
   *   The options array to be passed to the iterator client for the list
   *   request.
   */
  protected function buildListRequestOptions(
    array $filters,
    array $options
  ): array {
    // Support delay between requests that can reduce the risk of hitting
    // API rate and processing limits. If not given, no delay will be applied.
    $request_options = [
      'delay' => $options['delay'] ?? NULL,
    ];

    // Number of items per page. If not given, the iterator client's `list`
    // method will apply the default of 100.
    if (isset($options['limit'])) {
      $request_options['limit'] = $options['limit'];
    }
    // Page to start at; we may be continuing a state-managed import that did
    // not finish all pages in its last run.
    // If not given, the iterator client's `list` method will apply the default
    // of 1 i.e. start at the first page.
    if (!empty($filters['offset']) && isset($request_options['limit'])) {
      $request_options['page'] = ((int) floor($filters['offset'] / $request_options['limit'])) + 1;
    }

    return $request_options;
  }

}
