<?php

declare(strict_types=1);

namespace Drupal\entity_sync_odata\EntitySync\Api;

// Drupal modules.
use Drupal\entity_sync\Client\ProviderClientFactoryInterface;
// Drupal core.
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
// Third-party libraries.
use KrystalCode\Api\Session\SessionManagerInterface;

/**
 * Factory for generating OData v4 entity API clients.
 *
 * The clients generated by this factory require a session manager provided by
 * Entity Synchronization Session (`entity_sync_session`) included as a
 * submodule in the Entity Sychronization module.
 *
 * We do not add it as a dependency in this module because we may add here
 * another client for integrations that do not require authentication and
 * session management.
 *
 * Supported parameters:
 * - session_type: (string, required) The ID of the session type to use.
 * - entity_type: (string, required) The remote entity type for which to
 *   generate the client.
 * - session_duration: (int, optional) The minimum session duration to pass to
 *   the session manager, in seconds.
 * - session_duration (array, optional): An associative array containing
 *   information that instructs the session manager to ensure that sessions will
 *   have a minimum duration. That is, any authentication credentials e.g.
 *   access token, will be valid for at least the requested time. This is meant
 *   to be used, for example, in cases where several API calls will be made over
 *   a longer session and we want to minimize refreshing or recreating sessions
 *   to avoid exceeding session limits. Supported array items are:
 *   - interval (int, required): The minimum duration of the session.
 *   - limit (int, optional): The maximum duration of the session supported
 *     by the API provider, if known.
 * - odata: (array, optional) An array containing `select` and `expand` items to
 *   add to the corresponding OData clauses. Currenlty only supported by the
 *   `importList` client method. OData filters are not supported yet.
 * - client_class: (string, optional) The fully qualified name of the client
 *   class to instantiate. This may be useful in case it is necessary to
 *   extend/override the client provided by this module; it allows doing so
 *   without having to write a factory as well. The client class must be
 *   compatible with the instantiation done in this factory. Defaults to
 *   `\Drupal\entity_sync_odata\EntitySync\Api\EntityClient`.
 * - iterator_client_class: (string, optional) The fully qualified name of the
 *   iterator client to instantiate. Defaults to
 *   `\Drupal\entity_sync_odata\EntitySync\Api\SkipIteratorClient`.
 *
 * @see \Drupal\entity_sync_odata\EntitySync\Api\EntityClient::__construct()
 * @see \KrystalCode\Api\Session\SessionManagerInterface::setOptions()
 * phpcs:disable
 * @I Add client adapter for integrations that do not require authentication
 *    type     : feature
 *    priority : normal
 *    labels   : client
 * phpcs:enable
 */
class EntityClientFactory implements ProviderClientFactoryInterface {

  /**
   * Constructs a new EntityClientFactory object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The Drupal system time service.
   */
  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected TimeInterface $time
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public function get($sync_id, array $parameters) {
    $required_parameters = [
      'session_type',
      'entity_type',
    ];
    foreach ($required_parameters as $required_parameter) {
      if (!in_array($required_parameter, array_keys($parameters))) {
        throw new \InvalidArgumentException(sprintf(
          'Parameter "%s" is required for creating the OData entity client.',
          $required_parameter
        ));
      }
    }

    $session_manager = $this->getSessionManager($parameters);

    $iterator_client_class = SkipIteratorClient::class;
    if (isset($parameters['iterator_client_class'])) {
      $iterator_client_class = $parameters['iterator_client_class'];
    }

    $client_class = EntityClient::class;
    if (isset($parameters['client_class'])) {
      $client_class = $parameters['client_class'];
    }

    return new $client_class(
      $session_manager,
      new $iterator_client_class($session_manager, $parameters['entity_type']),
      $parameters['entity_type'],
      $parameters['odata'] ?? []
    );
  }

  /**
   * Instantiates, configures and returns the session manager.
   *
   * @param array $parameters
   *   An array containing the parameters provided by the synchronization.
   *
   * @return \KrystalCode\Api\Session\SessionManagerInterface
   *   The session manager.
   */
  protected function getSessionManager(
    array $parameters
  ): SessionManagerInterface {
    $session_type = $this->entityTypeManager
      ->getStorage('entity_sync_session_type')
      ->loadWithPluginInstantiated($parameters['session_type']);
    $session_manager = $session_type->getPlugin();

    $options = [];
    if (isset($parameters['session_duration'])) {
      $options['duration'] = $parameters['session_duration'];
      $options['duration']['start'] = $this->time->getRequestTime();
    }
    $options += $session_manager->getOptions();
    $session_manager->setOptions($options);

    return $session_manager;
  }

}
