<?php

declare(strict_types=1);

namespace Drupal\entity_sync_odata\EntitySync\Api;

// Third-party libraries.
use GuzzleHttp\HandlerStack;
use GuzzleHttp\Middleware;
use KrystalCode\Api\Session\SessionManagerInterface;
use KrystalCode\ApiIterator\ClientInterface as IteratorClientInterface;
use KrystalCode\ApiIterator\Iterator as ApiIterator;
use Psr\Http\Message\RequestInterface;
use SaintSystems\OData\Entity;

/**
 * Iterator client that uses the `Prefer` header and the skip token for paging.
 */
class SkipTokenIteratorClient implements IteratorClientInterface {

  use IteratorClientTrait;
  use ODataClientBuilderTrait;

  /**
   * Constructs a new PreferHeaderIteratorClient object.
   *
   * @param \KrystalCode\Api\Session\SessionManagerInterface $sessionManager
   *   The session manager.
   * @param string $entityType
   *   The entity type (resource) of the API client that will be used
   *   e.g. `People`.
   */
  public function __construct(
    protected SessionManagerInterface $sessionManager,
    protected string $entityType
  ) {
  }

  /**
   * {@inheritdoc}
   */
  public function list(array $options = [], array $query = []) {
    $options = array_merge(
      [
        'page' => 1,
        'limit' => 100,
        'delay' => NULL,
      ],
      $options
    );
    $query = array_merge(
      [
        'select' => NULL,
        'filter' => NULL,
        'expand' => NULL,
      ],
      $query
    );
    $query['skip'] = ($options['page'] - 1) * $options['limit'];
    $query['top'] = $options['limit'];

    // By default, we return an iterator. Caller code can then loop over it
    // and the iterator will internally handle making the API calls for
    // fetching the objects for each page.
    if (empty($options['bypass_iterator'])) {
      return new ApiIterator(
        $this,
        $options['page'],
        $options['limit'],
        $query,
        FALSE,
        $options['delay']
      );
    }

    // Otherwise, either it is requested to fetch the results directly, or
    // this is a call by the iterator while being looped over. Make the
    // request and return the results.
    unset($options['bypass_iterator']);

    [$objects, $query['skip_token']] = $this->getList($query);

    return [
      new \CachingIterator(
        $objects->getIterator(),
        \CachingIterator::FULL_CACHE
      ),
      // If we do not have a skip token in the response, we must be on the last
      // page.
      empty($query['skip_token']) ? FALSE : TRUE,
      $query,
    ];
  }

  /**
   * Fetches and returns the objects from the OData API.
   *
   * @param array $query
   *   The query parameters for fetching the objects.
   *
   * @return \Illuminate\Support\Collection
   *   The objects.
   */
  protected function getList(array $query): array {
    $client_config = [];
    if (isset($query['top'])) {
      $top = "odata.maxpagesize={$query['top']}";

      $stack = HandlerStack::create();
      $stack->push(Middleware::mapRequest(
        function (RequestInterface $request) use ($top) {
          return $request->withHeader('Prefer', $top);
        })
      );
      $client_config['handler'] = $stack;
    }

    $client = $this->getOdataClient($client_config);
    // Instruct the client to return the full response so that we can get the
    // skip token from it.
    $client->setEntityReturnType(FALSE);
    // Override the grammar to use our own until a bug is fixed in the default
    // grammar.
    $client->setQueryGrammar(new Grammar());
    $builder = $client->from($this->entityType);

    if (isset($query['skip_token'])) {
      $builder->skipToken($query['skip_token']);
    }

    if (isset($query['select'])) {
      $builder->select($query['select']);
    }
    if (isset($query['expand'])) {
      $builder->expand($query['expand']);
    }

    $builder = $this->buildFilter(
      $builder,
      $query['filter'][0],
      $query['filter'][1]
    );

    $response = $builder->get()->first();

    return [
      collect($response->getResponseAsObject(Entity::class)),
      $response->getSkipToken(),
    ];
  }

}
